(ns com.aurellem.gb.pokemon
  (:use (com.aurellem.gb gb-driver util constants characters))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(def pokemon-names-start 0xD2B4)

(defn set-party-number
  ([^SaveState state new-party-num]
     (set-memory state 0xD162 new-party-num))
  ([new-party-num]
     (set-party-number @current-state new-party-num)))

(def party-number-address 0xD162)

(defn party-number
  ([^SaveState state]
     (aget (memory state) party-number-address))
  ([] (party-number @current-state)))

(defn party-names
  ([^SaveState state]
     (let [raw-names
           (subvec (vec (memory state))
                   pokemon-names-start
                   (+ pokemon-names-start
                      (* name-width 6)))]
       (map
        read-name
        (take
         (party-number state)
          (partition name-width
                   raw-names)))))
  ([] (party-names @current-state)))

(defn rename-pokemon
  ([^SaveState state n new-name]
     (assert (<= 0 n (dec (party-number state))))
     (assert (<= (count new-name) max-name-length))
     (set-memory-range
      state
      (+ (* n name-width) pokemon-names-start)
      (concat (str->character-codes new-name) [end-of-name-marker])))
  ([n new-name]
     (rename-pokemon @current-state n new-name)))

(def OT-start 0xD272)

(defn original-trainers
  ([^SaveState state]
     (let [raw-names
           (subvec (vec (memory state))
                   OT-start
                   (+ OT-start
                      (* name-width 6)))]
       (map read-name
            (take (party-number state)
                  (partition name-width raw-names)))))
  ([] (original-trainers @current-state)))

(defn set-original-trainer
  "Set the OT name for a pokemon.
   Note that a pokemon is still considered 'yours' if
   the OT ID is the same as your own."
  ([^SaveState state poke-num new-name]
     (assert (<= 0 poke-num (dec (party-number state))))
     (assert (<= (count new-name) max-name-length))
     (set-memory-range
      state
      (+ (* poke-num name-width) OT-start)
      (concat (str->character-codes new-name) [end-of-name-marker])))
  ([n new-name]
     (set-original-trainer @current-state n new-name)))

(def OT-ID-addresses [0xD176 0xD1A2 0xD1CE 0xD1FA 0xD228 0xD252])

(defn set-pokemon-id
  ([^SaveState state n new-id]
     (assert (<= 0 n (dec (party-number state))))
     (assert (<= 0 new-id 0xFFFF))
     (set-memory-range 
      state
      (OT-ID-addresses n)
      [(bit-shift-right (bit-and new-id 0xFF00) 8)
       (bit-and new-id 0xFF)
       ]))
  ([n new-id]
     (set-pokemon-id @current-state n new-id)))

(def unknown "[[[UNKNOWN]]]")

(def unknown "")

(def pokemon-1-info
  {0xD16A "Color Map"            ;; 0
   0xD16B "Current-HP (h)"       ;; 1
   0xD16C "Current-HP (l)"       ;; 2
   0XD16D unknown                ;; 3
   0xD16E "Status"               ;; 4
   0xD16F unknown                ;; 5
   0xD170 unknown                ;; 6
   0xD171 unknown                ;; 7
   0xD172 "Move 1 ID"            ;; 8
   0xD173 "Move 2 ID"            ;; 9
   0xD174 "Move 3 ID"            ;; 10
   0xD175 "Move 4 ID"            ;; 11
   0xD176 "OT-ID (h)"            ;; 12
   0xD177 "OT-ID (l)"            ;; 13
   0xD178 "Exp. Points (h)"      ;; 14
   0xD179 "Exp. Points (m)"      ;; 15
   0xD17A "Exp. Points (l)"      ;; 16
   0xD17B "HP Exp. (h)"          ;; 17
   0xD17C "HP Exp. (l)"          ;; 18
   0xD17D "Attack Exp. (h)"      ;; 19
   0xD17E "Attack Exp. (l)"      ;; 20
   0xD17F "Defense Exp. (h)"     ;; 21
   0xD180 "Defense Exp. (l)"     ;; 22
   0xD181 "Speed Exp. (h)"       ;; 23
   0xD182 "Speed Exp. (l)"       ;; 24
   0xD183 "Special Exp. (h)"     ;; 25
   0xD184 "Special Exp. (l)"     ;; 26
   0xD185 "DV Atk/Def"           ;; 27
   0xD186 "DV Speed/Spc"         ;; 28
   0xD187 "PP Move 1"            ;; 29
   0xD188 "PP Move 2"            ;; 30
   0xD189 "PP Move 3"            ;; 31
   0xD18A "PP Move 4"            ;; 32
   0xD18B "Current Level"        ;; 33
   0xD18C "HP Total (h)"         ;; 34
   0xD18D "HP Total (l)"         ;; 35
   0xD18E "Attack   (h)"         ;; 36
   0xD18F "Attack   (l)"         ;; 37
   0xD190 "Defense  (h)"         ;; 38
   0xD191 "Defense  (l)"         ;; 39
   0xD192 "Speed    (h)"         ;; 40
   0xD193 "Speed    (l)"         ;; 41
   0xD194 "Special  (h)"         ;; 42
   0xD195 "Special  (l)"         ;; 43
   })

(defn pokemon-data
  ([^SaveState state pokemon-num]
     (assert (<= 0 pokemon-num 5))
     (let [base (+ (* pokemon-num  pokemon-record-width) 0xD16A)]
       (subvec (vec (memory state)) base
               (+ base pokemon-record-width))))
  ([pokemon-num] (pokemon-data @current-state pokemon-num)))

(defn set-pokemon-data
  ([^SaveState state pokemon-num new-data]
     (assert (<= 0 pokemon-num 5))
     (let [base (+ (* pokemon-num  pokemon-record-width) 0xD16A)]
       (set-memory-range state base new-data)))
  ([pokemon-num new-data]
     (set-pokemon-data @current-state pokemon-num new-data)))

(defn print-pokemon-data
  ([^SaveState state pokemon-num]
     (assert (<= 0 pokemon-num 5))
     (let [poke-data (pokemon-data state pokemon-num)
           backbone (sort (keys pokemon-1-info))]
       (println "Pokemon " pokemon-num  " -- "
                (nth (party-names state)
                     pokemon-num) \newline)

       (println "    Desc.          | Hex  | Dec |  Binary  |")
       (println "-------------------+------+-----+----------+")
       (dorun
        (map
         (comp println
               (fn [desc data]
                 (format "%-16s   | 0x%02X | %3d | %s |"
                         desc data data
                         (let [s (Integer/toBinaryString data)]
                           (apply
                            str 
                            (concat (repeat (- 8 (count s)) "0" )
                                    s))))))
         (map pokemon-1-info backbone)
         poke-data))))
  ([pokemon-num]
     (print-pokemon-data @current-state pokemon-num)))
